=== modified file 'bzrlib/bundle/serializer/v4.py'
--- bzrlib/bundle/serializer/v4.py	2008-06-18 07:56:09 +0000
+++ bzrlib/bundle/serializer/v4.py	2008-07-01 03:38:35 +0000
@@ -354,7 +354,8 @@
         mpdiffs = vf.make_mpdiffs(ordered_keys)
         sha1s = vf.get_sha1s(ordered_keys)
         parent_map = vf.get_parent_map(ordered_keys)
-        for mpdiff, item_key, sha1, in zip(mpdiffs, ordered_keys, sha1s):
+        for mpdiff, item_key, in zip(mpdiffs, ordered_keys):
+            sha1 = sha1s[item_key]
             parents = [key[-1] for key in parent_map[item_key]]
             text = ''.join(mpdiff.to_patch())
             # Infer file id records as appropriate.

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2008-06-11 04:20:16 +0000
+++ bzrlib/inventory.py	2008-06-19 13:14:24 +0000
@@ -472,9 +472,9 @@
 
     def _check(self, checker, tree_revision_id, tree):
         """See InventoryEntry._check"""
-        t = (self.file_id, self.revision)
-        if t in checker.checked_texts:
-            prev_sha = checker.checked_texts[t]
+        key = (self.file_id, self.revision)
+        if key in checker.checked_texts:
+            prev_sha = checker.checked_texts[key]
             if prev_sha != self.text_sha1:
                 raise BzrCheckError(
                     'mismatched sha1 on {%s} in {%s} (%s != %s) %r' %
@@ -489,11 +489,9 @@
         # We can't check the length, because Weave doesn't store that
         # information, and the whole point of looking at the weave's
         # sha1sum is that we don't have to extract the text.
-        if (self.text_sha1 !=
-            tree._repository.texts.get_sha1s([(self.file_id, self.revision)])[0]):
-            raise BzrCheckError('text {%s} version {%s} wrong sha1' 
-                                % (self.file_id, self.revision))
-        checker.checked_texts[t] = self.text_sha1
+        if (self.text_sha1 != tree._repository.texts.get_sha1s([key])[key]):
+            raise BzrCheckError('text {%s} version {%s} wrong sha1' % key)
+        checker.checked_texts[key] = self.text_sha1
 
     def copy(self):
         other = InventoryFile(self.file_id, self.name, self.parent_id)

=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2008-06-24 03:04:43 +0000
+++ bzrlib/knit.py	2008-07-01 03:49:54 +0000
@@ -730,6 +730,14 @@
             self._factory = KnitAnnotateFactory()
         else:
             self._factory = KnitPlainFactory()
+        self._fallback_vfs = []
+
+    def add_fallback_versioned_files(self, a_versioned_files):
+        """Add a source of texts for texts not present in this knit.
+
+        :param a_versioned_files: A VersionedFiles object.
+        """
+        self._fallback_vfs.append(a_versioned_files)
 
     def add_lines(self, key, parents, lines, parent_texts=None,
         left_matching_blocks=None, nostore_sha=None, random_id=False,
@@ -892,7 +900,11 @@
         fulltext_size = None
         for count in xrange(self._max_delta_chain):
             # XXX: Collapse these two queries:
-            method = self._index.get_method(parent)
+            try:
+                method = self._index.get_method(parent)
+            except RevisionNotPresent:
+                # Some basis is not locally present: always delta
+                return False
             index, pos, size = self._index.get_position(parent)
             if method == 'fulltext':
                 fulltext_size = size
@@ -961,9 +973,13 @@
         text_map, contents_map = self._get_content_maps([key])
         return contents_map[key]
 
-    def _get_content_maps(self, keys):
+    def _get_content_maps(self, keys, nonlocal_keys=None):
         """Produce maps of text and KnitContents
         
+        :param keys: The keys to produce content maps for.
+        :param nonlocal_keys: An iterable of keys(possibly intersecting keys)
+            which are known to not be in this knit, but rather in one of the
+            fallback knits.
         :return: (text_map, content_map) where text_map contains the texts for
             the requested versions and content_map contains the KnitContents.
         """
@@ -973,20 +989,46 @@
         # final output.
         keys = list(keys)
         multiple_versions = len(keys) != 1
-        record_map = self._get_record_map(keys)
+        record_map = self._get_record_map(keys, allow_missing=True)
 
         text_map = {}
         content_map = {}
         final_content = {}
+        if nonlocal_keys is None:
+            nonlocal_keys = set()
+        else:
+            nonlocal_keys = frozenset(nonlocal_keys)
+        missing_keys = set(nonlocal_keys)
+        for source in self._fallback_vfs:
+            if not missing_keys:
+                break
+            for record in source.get_record_stream(missing_keys,
+                'unordered', True):
+                if record.storage_kind == 'absent':
+                    continue
+                missing_keys.remove(record.key)
+                lines = split_lines(record.get_bytes_as('fulltext'))
+                text_map[record.key] = lines
+                content_map[record.key] = PlainKnitContent(lines, record.key)
+                if record.key in keys:
+                    final_content[record.key] = content_map[record.key]
         for key in keys:
+            if key in nonlocal_keys:
+                # already handled
+                continue
             components = []
             cursor = key
             while cursor is not None:
-                record, record_details, digest, next = record_map[cursor]
+                try:
+                    record, record_details, digest, next = record_map[cursor]
+                except KeyError:
+                    raise RevisionNotPresent(cursor, self)
                 components.append((cursor, record, record_details, digest))
+                cursor = next
                 if cursor in content_map:
+                    # no need to plan further back
+                    components.append((cursor, None, None, None))
                     break
-                cursor = next
 
             content = None
             for (component_id, record, record_details,
@@ -1022,9 +1064,32 @@
         :return: A mapping from keys to parents. Absent keys are absent from
             the mapping.
         """
-        return self._index.get_parent_map(keys)
-
-    def _get_record_map(self, keys):
+        return self._get_parent_map_with_sources(keys)[0]
+
+    def _get_parent_map_with_sources(self, keys):
+        """Get a map of the parents of keys.
+
+        :param keys: The keys to look up parents for.
+        :return: A tuple. The first element is a mapping from keys to parents.
+            Absent keys are absent from the mapping. The second element is a
+            list with the locations each key was found in. The first element
+            is the in-this-knit parents, the second the first fallback source,
+            and so on.
+        """
+        result = {}
+        sources = [self._index] + self._fallback_vfs
+        source_results = []
+        missing = set(keys)
+        for source in sources:
+            if not missing:
+                break
+            new_result = source.get_parent_map(missing)
+            source_results.append(new_result)
+            result.update(new_result)
+            missing.difference_update(set(new_result))
+        return result, source_results
+
+    def _get_record_map(self, keys, allow_missing=False):
         """Produce a dictionary of knit records.
         
         :return: {key:(record, record_details, digest, next)}
@@ -1037,8 +1102,12 @@
             next
                 build-parent of the version, i.e. the leftmost ancestor.
                 Will be None if the record is not a delta.
+        :param keys: The keys to build a map for
+        :param allow_missing: If some records are missing, rather than 
+            error, just return the data that could be generated.
         """
-        position_map = self._get_components_positions(keys)
+        position_map = self._get_components_positions(keys,
+            allow_missing=allow_missing)
         # key = component_id, r = record_details, i_m = index_memo, n = next
         records = [(key, i_m) for key, (r, i_m, n)
                              in position_map.iteritems()]
@@ -1063,6 +1132,8 @@
         """
         # keys might be a generator
         keys = set(keys)
+        if not keys:
+            return
         if not self._index.has_graph:
             # Cannot topological order when no graph has been stored.
             ordering = 'unordered'
@@ -1078,6 +1149,7 @@
         # There may be more absent keys : if we're missing the basis component
         # and are trying to include the delta closure.
         if include_delta_closure:
+            needed_from_fallback = set()
             # Build up reconstructable_keys dict.  key:True in this dict means
             # the key can be reconstructed.
             reconstructable_keys = {}
@@ -1086,7 +1158,7 @@
                 try:
                     chain = [key, positions[key][2]]
                 except KeyError:
-                    absent_keys.add(key)
+                    needed_from_fallback.add(key)
                     continue
                 result = True
                 while chain[-1] is not None:
@@ -1098,44 +1170,87 @@
                             chain.append(positions[chain[-1]][2])
                         except KeyError:
                             # missing basis component
-                            result = False
+                            needed_from_fallback.add(chain[-1])
+                            result = True
                             break
                 for chain_key in chain[:-1]:
                     reconstructable_keys[chain_key] = result
                 if not result:
-                    absent_keys.add(key)
+                    needed_from_fallback.add(key)
+        # Double index lookups here : need a unified api ?
+        global_map, parent_maps = self._get_parent_map_with_sources(keys)
+        if ordering == 'topological':
+            # Global topological sort
+            present_keys = topo_sort(global_map)
+            # Now group by source:
+            source_keys = []
+            current_source = None
+            for key in present_keys:
+                for parent_map in parent_maps:
+                    if key in parent_map:
+                        key_source = parent_map
+                        break
+                if current_source is not key_source:
+                    source_keys.append((key_source, []))
+                    current_source = key_source
+                source_keys[-1][1].append(key)
+        else:
+            # Just group by source; remote sources first.
+            present_keys = []
+            source_keys = []
+            for parent_map in reversed(parent_maps):
+                source_keys.append((parent_map, []))
+                for key in parent_map:
+                    present_keys.append(key)
+                    source_keys[-1][1].append(key)
+        absent_keys = keys - set(global_map)
         for key in absent_keys:
             yield AbsentContentFactory(key)
         # restrict our view to the keys we can answer.
-        keys = keys - absent_keys
-        # Double index lookups here : need a unified api ?
-        parent_map = self.get_parent_map(keys)
-        if ordering == 'topological':
-            present_keys = topo_sort(parent_map)
-        else:
-            present_keys = keys
         # XXX: Memory: TODO: batch data here to cap buffered data at (say) 1MB.
-        # XXX: At that point we need to consider double reads by utilising
-        # components multiple times.
+        # XXX: At that point we need to consider the impact of double reads by
+        # utilising components multiple times.
         if include_delta_closure:
             # XXX: get_content_maps performs its own index queries; allow state
             # to be passed in.
-            text_map, _ = self._get_content_maps(present_keys)
+            text_map, _ = self._get_content_maps(present_keys,
+                needed_from_fallback - absent_keys)
             for key in present_keys:
-                yield FulltextContentFactory(key, parent_map[key], None,
+                yield FulltextContentFactory(key, global_map[key], None,
                     ''.join(text_map[key]))
         else:
-            records = [(key, positions[key][1]) for key in present_keys]
-            for key, raw_data, sha1 in self._read_records_iter_raw(records):
-                (record_details, index_memo, _) = positions[key]
-                yield KnitContentFactory(key, parent_map[key],
-                    record_details, sha1, raw_data, self._factory.annotated, None)
+            for source, keys in source_keys:
+                if source is parent_maps[0]:
+                    # this KnitVersionedFiles
+                    records = [(key, positions[key][1]) for key in keys]
+                    for key, raw_data, sha1 in self._read_records_iter_raw(records):
+                        (record_details, index_memo, _) = positions[key]
+                        yield KnitContentFactory(key, global_map[key],
+                            record_details, sha1, raw_data, self._factory.annotated, None)
+                else:
+                    vf = self._fallback_vfs[parent_maps.index(source) - 1]
+                    for record in vf.get_record_stream(keys, ordering,
+                        include_delta_closure):
+                        yield record
 
     def get_sha1s(self, keys):
         """See VersionedFiles.get_sha1s()."""
-        record_map = self._get_record_map(keys)
-        # record entry 2 is the 'digest'.
-        return [record_map[key][2] for key in keys]
+        missing = set(keys)
+        record_map = self._get_record_map(missing, allow_missing=True)
+        result = {}
+        for key, details in record_map.iteritems():
+            if key not in missing:
+                continue
+            # record entry 2 is the 'digest'.
+            result[key] = details[2]
+        missing.difference_update(set(result))
+        for source in self._fallback_vfs:
+            if not missing:
+                break
+            new_result = source.get_sha1s(missing)
+            result.update(new_result)
+            missing.difference_update(set(new_result))
+        return result
 
     def insert_record_stream(self, stream):
         """Insert a record stream into this container.
@@ -1274,20 +1389,16 @@
         if pb is None:
             pb = progress.DummyProgress()
         keys = set(keys)
-        # filter for available keys
-        parent_map = self.get_parent_map(keys)
-        if len(parent_map) != len(keys):
-            missing = set(parent_map) - requested_keys
-            raise RevisionNotPresent(key, self.filename)
+        total = len(keys)
         # we don't care about inclusions, the caller cares.
         # but we need to setup a list of records to visit.
         # we need key, position, length
         key_records = []
         build_details = self._index.get_build_details(keys)
-        for key in keys:
-            # [0] is index_memo
-            key_records.append((key, build_details[key][0]))
-        total = len(key_records)
+        for key, details in build_details.iteritems():
+            if key in keys:
+                key_records.append((key, details[0]))
+                keys.remove(key)
         records_iter = enumerate(self._read_records_iter(key_records))
         for (key_idx, (key, data, sha_value)) in records_iter:
             pb.update('Walking content.', key_idx, total)
@@ -1303,6 +1414,16 @@
             # change to integrate into the rest of the codebase. RBC 20071110
             for line in line_iterator:
                 yield line, key
+        for source in self._fallback_vfs:
+            if not keys:
+                break
+            source_keys = set()
+            for line, key in source.iter_lines_added_or_present_in_keys(keys):
+                source_keys.add(key)
+                yield line, key
+            keys.difference_update(source_keys)
+        if keys:
+            raise RevisionNotPresent(keys, self.filename)
         pb.update('Walking content.', total, total)
 
     def _make_line_delta(self, delta_seq, new_content):
@@ -1504,7 +1625,12 @@
         """See VersionedFiles.keys."""
         if 'evil' in debug.debug_flags:
             trace.mutter_callsite(2, "keys scales with size of history")
-        return self._index.keys()
+        sources = [self._index] + self._fallback_vfs
+        result = set()
+        for source in sources:
+            result.update(source.keys())
+        return result
+
 
 
 class _KndxIndex(object):
@@ -1730,7 +1856,10 @@
         """
         prefix, suffix = self._split_key(key)
         self._load_prefixes([prefix])
-        return self._kndx_cache[prefix][0][suffix][1]
+        try:
+            return self._kndx_cache[prefix][0][suffix][1]
+        except KeyError:
+            raise RevisionNotPresent(key, self)
 
     def get_parent_map(self, keys):
         """Get a map of the parents of keys.

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-06-26 00:17:06 +0000
+++ bzrlib/repository.py	2008-07-01 03:38:35 +0000
@@ -584,7 +584,8 @@
                 rev.inventory_sha1 = self.add_inventory(revision_id, inv,
                                                         rev.parent_ids)
         else:
-            rev.inventory_sha1 = self.inventories.get_sha1s([(revision_id,)])[0]
+            key = (revision_id,)
+            rev.inventory_sha1 = self.inventories.get_sha1s([key])[key]
         self._add_revision(rev)
 
     def _add_revision(self, revision):

=== modified file 'bzrlib/tests/repository_implementations/test_check_reconcile.py'
--- bzrlib/tests/repository_implementations/test_check_reconcile.py	2008-06-11 04:20:16 +0000
+++ bzrlib/tests/repository_implementations/test_check_reconcile.py	2008-06-19 13:14:24 +0000
@@ -150,7 +150,7 @@
         :returns: A dict of `{version: hash}`.
         """
         keys = [('a-file-id', version) for version in versions]
-        return dict(zip(versions, repo.texts.get_sha1s(keys)))
+        return repo.texts.get_sha1s(keys)
 
     def test_reconcile_behaviour(self):
         """Populate a repository and reconcile it, verifying the state before

=== modified file 'bzrlib/tests/test_knit.py'
--- bzrlib/tests/test_knit.py	2008-06-11 04:20:16 +0000
+++ bzrlib/tests/test_knit.py	2008-07-01 03:49:54 +0000
@@ -26,6 +26,7 @@
     errors,
     generate_ids,
     knit,
+    multiparent,
     pack,
     )
 from bzrlib.errors import (
@@ -51,6 +52,7 @@
 from bzrlib.symbol_versioning import one_four
 from bzrlib.tests import (
     Feature,
+    KnownFailure,
     TestCase,
     TestCaseWithMemoryTransport,
     TestCaseWithTransport,
@@ -58,7 +60,11 @@
 from bzrlib.transport import get_transport
 from bzrlib.transport.memory import MemoryTransport
 from bzrlib.tuned_gzip import GzipFile
-from bzrlib.versionedfile import ConstantMapper
+from bzrlib.versionedfile import (
+    AbsentContentFactory,
+    ConstantMapper,
+    RecordingVersionedFilesDecorator,
+    )
 
 
 class _CompiledKnitFeature(Feature):
@@ -1371,3 +1377,398 @@
             [(('tip',), 'fulltext,no-eol', (None, 0, 100), []),
              (('tip',), 'no-eol,line-delta', (None, 0, 100), [])])
         self.assertEqual([], self.caught_entries)
+
+
+class TestStacking(KnitTests):
+
+    def get_basis_and_test_knit(self):
+        basis = self.make_test_knit(name='basis')
+        basis = RecordingVersionedFilesDecorator(basis)
+        test = self.make_test_knit(name='test')
+        test.add_fallback_versioned_files(basis)
+        return basis, test
+
+    def test_add_fallback_versioned_files(self):
+        basis = self.make_test_knit(name='basis')
+        test = self.make_test_knit(name='test')
+        # It must not error; other tests test that the fallback is referred to
+        # when accessing data.
+        test.add_fallback_versioned_files(basis)
+
+    def test_add_lines(self):
+        # lines added to the test are not added to the basis
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_cross_border = ('quux',)
+        key_delta = ('zaphod',)
+        test.add_lines(key, (), ['foo\n'])
+        self.assertEqual({}, basis.get_parent_map([key]))
+        # lines added to the test that reference across the stack do a
+        # fulltext.
+        basis.add_lines(key_basis, (), ['foo\n'])
+        basis.calls = []
+        test.add_lines(key_cross_border, (key_basis,), ['foo\n'])
+        self.assertEqual('fulltext', test._index.get_method(key_cross_border))
+        self.assertEqual([("get_parent_map", set([key_basis]))], basis.calls)
+        # Subsequent adds do delta.
+        basis.calls = []
+        test.add_lines(key_delta, (key_cross_border,), ['foo\n'])
+        self.assertEqual('line-delta', test._index.get_method(key_delta))
+        self.assertEqual([], basis.calls)
+
+    def test_annotate(self):
+        # annotations from the test knit are answered without asking the basis
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_missing = ('missing',)
+        test.add_lines(key, (), ['foo\n'])
+        details = test.annotate(key)
+        self.assertEqual([(key, 'foo\n')], details)
+        self.assertEqual([], basis.calls)
+        # But texts that are not in the test knit are looked for in the basis
+        # directly.
+        basis.add_lines(key_basis, (), ['foo\n', 'bar\n'])
+        basis.calls = []
+        self.assertRaises(RevisionNotPresent, test.annotate, key_basis)
+        raise KnownFailure("Annotation on stacked knits currently fails.")
+        details = test.annotate(key_basis)
+        self.assertEqual([(key_basis, 'foo\n'), (key_basis, 'bar\n')], details)
+        self.assertEqual([("annotate", key_basis)], basis.calls)
+
+    def test_check(self):
+        # check() must not check the fallback files, it's none of its business.
+        basis, test = self.get_basis_and_test_knit()
+        basis.check = None
+        test.check()
+
+    def test_get_parent_map(self):
+        # parents in the test knit are answered without asking the basis
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_missing = ('missing',)
+        test.add_lines(key, (), [])
+        parent_map = test.get_parent_map([key])
+        self.assertEqual({key: ()}, parent_map)
+        self.assertEqual([], basis.calls)
+        # But parents that are not in the test knit are looked for in the basis
+        basis.add_lines(key_basis, (), [])
+        basis.calls = []
+        parent_map = test.get_parent_map([key, key_basis, key_missing])
+        self.assertEqual({key: (),
+            key_basis: ()}, parent_map)
+        self.assertEqual([("get_parent_map", set([key_basis, key_missing]))],
+            basis.calls)
+
+    def test_get_record_stream_unordered_fulltexts(self):
+        # records from the test knit are answered without asking the basis:
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_missing = ('missing',)
+        test.add_lines(key, (), ['foo\n'])
+        records = list(test.get_record_stream([key], 'unordered', True))
+        self.assertEqual(1, len(records))
+        self.assertEqual([], basis.calls)
+        # Missing (from test knit) objects are retrieved from the basis:
+        basis.add_lines(key_basis, (), ['foo\n', 'bar\n'])
+        basis.calls = []
+        records = list(test.get_record_stream([key_basis, key_missing],
+            'unordered', True))
+        self.assertEqual(2, len(records))
+        calls = list(basis.calls)
+        for record in records:
+            self.assertSubset([record.key], (key_basis, key_missing))
+            if record.key == key_missing:
+                self.assertIsInstance(record, AbsentContentFactory)
+            else:
+                reference = list(basis.get_record_stream([key_basis],
+                    'unordered', True))[0]
+                self.assertEqual(reference.key, record.key)
+                self.assertEqual(reference.sha1, record.sha1)
+                self.assertEqual(reference.storage_kind, record.storage_kind)
+                self.assertEqual(reference.get_bytes_as(reference.storage_kind),
+                    record.get_bytes_as(record.storage_kind))
+                self.assertEqual(reference.get_bytes_as('fulltext'),
+                    record.get_bytes_as('fulltext'))
+        # It's not strictly minimal, but it seems reasonable for now for it to
+        # ask which fallbacks have which parents.
+        self.assertEqual([
+            ("get_parent_map", set([key_basis, key_missing])),
+            ("get_record_stream", [key_basis], 'unordered', True)],
+            calls)
+
+    def test_get_record_stream_ordered_fulltexts(self):
+        # ordering is preserved down into the fallback store.
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_basis_2 = ('quux',)
+        key_missing = ('missing',)
+        test.add_lines(key, (key_basis,), ['foo\n'])
+        # Missing (from test knit) objects are retrieved from the basis:
+        basis.add_lines(key_basis, (key_basis_2,), ['foo\n', 'bar\n'])
+        basis.add_lines(key_basis_2, (), ['quux\n'])
+        basis.calls = []
+        # ask for in non-topological order
+        records = list(test.get_record_stream(
+            [key, key_basis, key_missing, key_basis_2], 'topological', True))
+        self.assertEqual(4, len(records))
+        results = []
+        for record in records:
+            self.assertSubset([record.key],
+                (key_basis, key_missing, key_basis_2, key))
+            if record.key == key_missing:
+                self.assertIsInstance(record, AbsentContentFactory)
+            else:
+                results.append((record.key, record.sha1, record.storage_kind,
+                    record.get_bytes_as('fulltext')))
+        calls = list(basis.calls)
+        order = [record[0] for record in results]
+        self.assertEqual([key_basis_2, key_basis, key], order)
+        for result in results:
+            if result[0] == key:
+                source = test
+            else:
+                source = basis
+            record = source.get_record_stream([result[0]], 'unordered',
+                True).next()
+            self.assertEqual(record.key, result[0])
+            self.assertEqual(record.sha1, result[1])
+            self.assertEqual(record.storage_kind, result[2])
+            self.assertEqual(record.get_bytes_as('fulltext'), result[3])
+        # It's not strictly minimal, but it seems reasonable for now for it to
+        # ask which fallbacks have which parents.
+        self.assertEqual([
+            ("get_parent_map", set([key_basis, key_basis_2, key_missing])),
+            # unordered is asked for by the underlying worker as it still
+            # buffers everything while answering - which is a problem!
+            ("get_record_stream", [key_basis_2, key_basis], 'unordered', True)],
+            calls)
+
+    def test_get_record_stream_unordered_deltas(self):
+        # records from the test knit are answered without asking the basis:
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_missing = ('missing',)
+        test.add_lines(key, (), ['foo\n'])
+        records = list(test.get_record_stream([key], 'unordered', False))
+        self.assertEqual(1, len(records))
+        self.assertEqual([], basis.calls)
+        # Missing (from test knit) objects are retrieved from the basis:
+        basis.add_lines(key_basis, (), ['foo\n', 'bar\n'])
+        basis.calls = []
+        records = list(test.get_record_stream([key_basis, key_missing],
+            'unordered', False))
+        self.assertEqual(2, len(records))
+        calls = list(basis.calls)
+        for record in records:
+            self.assertSubset([record.key], (key_basis, key_missing))
+            if record.key == key_missing:
+                self.assertIsInstance(record, AbsentContentFactory)
+            else:
+                reference = list(basis.get_record_stream([key_basis],
+                    'unordered', False))[0]
+                self.assertEqual(reference.key, record.key)
+                self.assertEqual(reference.sha1, record.sha1)
+                self.assertEqual(reference.storage_kind, record.storage_kind)
+                self.assertEqual(reference.get_bytes_as(reference.storage_kind),
+                    record.get_bytes_as(record.storage_kind))
+        # It's not strictly minimal, but it seems reasonable for now for it to
+        # ask which fallbacks have which parents.
+        self.assertEqual([
+            ("get_parent_map", set([key_basis, key_missing])),
+            ("get_record_stream", [key_basis], 'unordered', False)],
+            calls)
+
+    def test_get_record_stream_ordered_deltas(self):
+        # ordering is preserved down into the fallback store.
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_basis_2 = ('quux',)
+        key_missing = ('missing',)
+        test.add_lines(key, (key_basis,), ['foo\n'])
+        # Missing (from test knit) objects are retrieved from the basis:
+        basis.add_lines(key_basis, (key_basis_2,), ['foo\n', 'bar\n'])
+        basis.add_lines(key_basis_2, (), ['quux\n'])
+        basis.calls = []
+        # ask for in non-topological order
+        records = list(test.get_record_stream(
+            [key, key_basis, key_missing, key_basis_2], 'topological', False))
+        self.assertEqual(4, len(records))
+        results = []
+        for record in records:
+            self.assertSubset([record.key],
+                (key_basis, key_missing, key_basis_2, key))
+            if record.key == key_missing:
+                self.assertIsInstance(record, AbsentContentFactory)
+            else:
+                results.append((record.key, record.sha1, record.storage_kind,
+                    record.get_bytes_as(record.storage_kind)))
+        calls = list(basis.calls)
+        order = [record[0] for record in results]
+        self.assertEqual([key_basis_2, key_basis, key], order)
+        for result in results:
+            if result[0] == key:
+                source = test
+            else:
+                source = basis
+            record = source.get_record_stream([result[0]], 'unordered',
+                False).next()
+            self.assertEqual(record.key, result[0])
+            self.assertEqual(record.sha1, result[1])
+            self.assertEqual(record.storage_kind, result[2])
+            self.assertEqual(record.get_bytes_as(record.storage_kind), result[3])
+        # It's not strictly minimal, but it seems reasonable for now for it to
+        # ask which fallbacks have which parents.
+        self.assertEqual([
+            ("get_parent_map", set([key_basis, key_basis_2, key_missing])),
+            ("get_record_stream", [key_basis_2, key_basis], 'topological', False)],
+            calls)
+
+    def test_get_sha1s(self):
+        # sha1's in the test knit are answered without asking the basis
+        basis, test = self.get_basis_and_test_knit()
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_missing = ('missing',)
+        test.add_lines(key, (), ['foo\n'])
+        key_sha1sum = sha.new('foo\n').hexdigest()
+        sha1s = test.get_sha1s([key])
+        self.assertEqual({key: key_sha1sum}, sha1s)
+        self.assertEqual([], basis.calls)
+        # But texts that are not in the test knit are looked for in the basis
+        # directly (rather than via text reconstruction) so that remote servers
+        # etc don't have to answer with full content.
+        basis.add_lines(key_basis, (), ['foo\n', 'bar\n'])
+        basis_sha1sum = sha.new('foo\nbar\n').hexdigest()
+        basis.calls = []
+        sha1s = test.get_sha1s([key, key_missing, key_basis])
+        self.assertEqual({key: key_sha1sum,
+            key_basis: basis_sha1sum}, sha1s)
+        self.assertEqual([("get_sha1s", set([key_basis, key_missing]))],
+            basis.calls)
+
+    def test_insert_record_stream(self):
+        # records are inserted as normal; insert_record_stream builds on
+        # add_lines, so a smoke test should be all that's needed:
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_delta = ('zaphod',)
+        basis, test = self.get_basis_and_test_knit()
+        source = self.make_test_knit(name='source')
+        basis.add_lines(key_basis, (), ['foo\n'])
+        basis.calls = []
+        source.add_lines(key_basis, (), ['foo\n'])
+        source.add_lines(key_delta, (key_basis,), ['bar\n'])
+        stream = source.get_record_stream([key_delta], 'unordered', False)
+        test.insert_record_stream(stream)
+        self.assertEqual([("get_parent_map", set([key_basis]))],
+            basis.calls)
+        self.assertEqual({key_delta:(key_basis,)},
+            test.get_parent_map([key_delta]))
+        self.assertEqual('bar\n', test.get_record_stream([key_delta],
+            'unordered', True).next().get_bytes_as('fulltext'))
+
+    def test_iter_lines_added_or_present_in_keys(self):
+        # Lines from the basis are returned, and lines for a given key are only
+        # returned once. 
+        key1 = ('foo1',)
+        key2 = ('foo2',)
+        # all sources are asked for keys:
+        basis, test = self.get_basis_and_test_knit()
+        basis.add_lines(key1, (), ["foo"])
+        basis.calls = []
+        lines = list(test.iter_lines_added_or_present_in_keys([key1]))
+        self.assertEqual([("foo\n", key1)], lines)
+        self.assertEqual([("iter_lines_added_or_present_in_keys", set([key1]))],
+            basis.calls)
+        # keys in both are not duplicated:
+        test.add_lines(key2, (), ["bar\n"])
+        basis.add_lines(key2, (), ["bar\n"])
+        basis.calls = []
+        lines = list(test.iter_lines_added_or_present_in_keys([key2]))
+        self.assertEqual([("bar\n", key2)], lines)
+        self.assertEqual([], basis.calls)
+
+    def test_keys(self):
+        key1 = ('foo1',)
+        key2 = ('foo2',)
+        # all sources are asked for keys:
+        basis, test = self.get_basis_and_test_knit()
+        keys = test.keys()
+        self.assertEqual(set(), set(keys))
+        self.assertEqual([("keys",)], basis.calls)
+        # keys from a basis are returned:
+        basis.add_lines(key1, (), [])
+        basis.calls = []
+        keys = test.keys()
+        self.assertEqual(set([key1]), set(keys))
+        self.assertEqual([("keys",)], basis.calls)
+        # keys in both are not duplicated:
+        test.add_lines(key2, (), [])
+        basis.add_lines(key2, (), [])
+        basis.calls = []
+        keys = test.keys()
+        self.assertEqual(2, len(keys))
+        self.assertEqual(set([key1, key2]), set(keys))
+        self.assertEqual([("keys",)], basis.calls)
+
+    def test_add_mpdiffs(self):
+        # records are inserted as normal; add_mpdiff builds on
+        # add_lines, so a smoke test should be all that's needed:
+        key = ('foo',)
+        key_basis = ('bar',)
+        key_delta = ('zaphod',)
+        basis, test = self.get_basis_and_test_knit()
+        source = self.make_test_knit(name='source')
+        basis.add_lines(key_basis, (), ['foo\n'])
+        basis.calls = []
+        source.add_lines(key_basis, (), ['foo\n'])
+        source.add_lines(key_delta, (key_basis,), ['bar\n'])
+        diffs = source.make_mpdiffs([key_delta])
+        test.add_mpdiffs([(key_delta, (key_basis,),
+            source.get_sha1s([key_delta])[key_delta], diffs[0])])
+        self.assertEqual([("get_parent_map", set([key_basis])),
+            ('get_record_stream', [key_basis], 'unordered', True),
+            ('get_parent_map', set([key_basis]))],
+            basis.calls)
+        self.assertEqual({key_delta:(key_basis,)},
+            test.get_parent_map([key_delta]))
+        self.assertEqual('bar\n', test.get_record_stream([key_delta],
+            'unordered', True).next().get_bytes_as('fulltext'))
+
+    def test_make_mpdiffs(self):
+        # Generating an mpdiff across a stacking boundary should detect parent
+        # texts regions.
+        key = ('foo',)
+        key_left = ('bar',)
+        key_right = ('zaphod',)
+        basis, test = self.get_basis_and_test_knit()
+        basis.add_lines(key_left, (), ['bar\n'])
+        basis.add_lines(key_right, (), ['zaphod\n'])
+        basis.calls = []
+        test.add_lines(key, (key_left, key_right),
+            ['bar\n', 'foo\n', 'zaphod\n'])
+        diffs = test.make_mpdiffs([key])
+        self.assertEqual([
+            multiparent.MultiParent([multiparent.ParentText(0, 0, 0, 1),
+                multiparent.NewText(['foo\n']),
+                multiparent.ParentText(1, 0, 2, 1)])],
+            diffs)
+        self.assertEqual(4, len(basis.calls))
+        self.assertEqual([
+            ("get_parent_map", set([key_left, key_right])),
+            ("get_parent_map", set([key_left, key_right])),
+            ("get_parent_map", set([key_left, key_right])),
+            ],
+            basis.calls[:3])
+        last_call = basis.calls[3]
+        self.assertEqual('get_record_stream', last_call[0])
+        self.assertEqual(set([key_left, key_right]), set(last_call[1]))
+        self.assertEqual('unordered', last_call[2])
+        self.assertEqual(True, last_call[3])

=== modified file 'bzrlib/tests/test_versionedfile.py'
--- bzrlib/tests/test_versionedfile.py	2008-06-11 07:22:00 +0000
+++ bzrlib/tests/test_versionedfile.py	2008-06-19 13:14:24 +0000
@@ -501,7 +501,7 @@
         for version in multiparent.topo_iter(vf):
             mpdiff = vf.make_mpdiffs([version])[0]
             new_vf.add_mpdiffs([(version, vf.get_parent_map([version])[version],
-                                 vf.get_sha1s([version])[0], mpdiff)])
+                                 vf.get_sha1s([version])[version], mpdiff)])
             self.assertEqualDiff(vf.get_text(version),
                                  new_vf.get_text(version))
 
@@ -840,10 +840,12 @@
         vf.add_lines('b', ['a'], ['a\n'])
         # a file differing only in last newline.
         vf.add_lines('c', [], ['a'])
-        self.assertEqual(['3f786850e387550fdab836ed7e6dc881de23001b',
-                          '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8',
-                          '3f786850e387550fdab836ed7e6dc881de23001b'],
-                          vf.get_sha1s(['a', 'c', 'b']))
+        self.assertEqual({
+            'a': '3f786850e387550fdab836ed7e6dc881de23001b',
+            'c': '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8',
+            'b': '3f786850e387550fdab836ed7e6dc881de23001b',
+            },
+            vf.get_sha1s(['a', 'c', 'b']))
         
 
 class TestWeave(TestCaseWithMemoryTransport, VersionedFileTestMixIn):
@@ -1562,7 +1564,8 @@
         for factory in entries:
             on_seen(factory.key)
             self.assertValidStorageKind(factory.storage_kind)
-            self.assertEqual(f.get_sha1s([factory.key])[0], factory.sha1)
+            self.assertEqual(f.get_sha1s([factory.key])[factory.key],
+                factory.sha1)
             self.assertEqual(parents[factory.key], factory.parents)
             self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
                 str)
@@ -1627,7 +1630,8 @@
         for factory in entries:
             seen.append(factory.key)
             self.assertValidStorageKind(factory.storage_kind)
-            self.assertSubset([factory.sha1], [None, files.get_sha1s([factory.key])[0]])
+            self.assertSubset([factory.sha1],
+                [None, files.get_sha1s([factory.key])[factory.key]])
             self.assertEqual(parent_map[factory.key], factory.parents)
             # self.assertEqual(files.get_text(factory.key),
             self.assertIsInstance(factory.get_bytes_as('fulltext'), str)
@@ -1671,7 +1675,8 @@
         for factory in entries:
             seen.add(factory.key)
             self.assertValidStorageKind(factory.storage_kind)
-            self.assertEqual(files.get_sha1s([factory.key])[0], factory.sha1)
+            self.assertEqual(files.get_sha1s([factory.key])[factory.key],
+                factory.sha1)
             self.assertEqual(parent_map[factory.key], factory.parents)
             # currently no stream emits mpdiff
             self.assertRaises(errors.UnavailableRepresentation,
@@ -1710,7 +1715,8 @@
                 self.assertEqual(None, factory.parents)
             else:
                 self.assertValidStorageKind(factory.storage_kind)
-                self.assertEqual(files.get_sha1s([factory.key])[0], factory.sha1)
+                self.assertEqual(files.get_sha1s([factory.key])[factory.key],
+                    factory.sha1)
                 self.assertEqual(parents[factory.key], factory.parents)
                 self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
                     str)
@@ -1798,13 +1804,13 @@
                 ('FileA', 'base'), ('FileB', 'origin'), ('FileA', 'left'),
                 ('FileA', 'merged'), ('FileB', 'right'),
                 ]
-        self.assertEqual([
-            '51c64a6f4fc375daf0d24aafbabe4d91b6f4bb44',
-            '00e364d235126be43292ab09cb4686cf703ddc17',
-            'a8478686da38e370e32e42e8a0c220e33ee9132f',
-            'ed8bce375198ea62444dc71952b22cfc2b09226d',
-            '9ef09dfa9d86780bdec9219a22560c6ece8e0ef1',
-            ],
+        self.assertEqual({
+            keys[0]: '51c64a6f4fc375daf0d24aafbabe4d91b6f4bb44',
+            keys[1]: '00e364d235126be43292ab09cb4686cf703ddc17',
+            keys[2]: 'a8478686da38e370e32e42e8a0c220e33ee9132f',
+            keys[3]: 'ed8bce375198ea62444dc71952b22cfc2b09226d',
+            keys[4]: '9ef09dfa9d86780bdec9219a22560c6ece8e0ef1',
+            },
             files.get_sha1s(keys))
         
     def test_insert_record_stream_empty(self):
@@ -2132,7 +2138,7 @@
             mpdiff = files.make_mpdiffs([key])[0]
             parents = files.get_parent_map([key])[key] or []
             target.add_mpdiffs(
-                [(key, parents, files.get_sha1s([key])[0], mpdiff)])
+                [(key, parents, files.get_sha1s([key])[key], mpdiff)])
             self.assertEqualDiff(
                 files.get_record_stream([key], 'unordered',
                     True).next().get_bytes_as('fulltext'),

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2008-06-18 07:56:09 +0000
+++ bzrlib/versionedfile.py	2008-07-01 03:49:54 +0000
@@ -19,6 +19,7 @@
 
 """Versioned text file storage api."""
 
+from copy import copy
 from cStringIO import StringIO
 import os
 import urllib
@@ -372,19 +373,11 @@
                     parent_ids, lines, vf_parents,
                     left_matching_blocks=left_matching_blocks)
             vf_parents[version] = version_text
-        for (version, parent_ids, expected_sha1, mpdiff), sha1 in\
-             zip(records, self.get_sha1s(versions)):
-            if expected_sha1 != sha1:
+        sha1s = self.get_sha1s(versions)
+        for version, parent_ids, expected_sha1, mpdiff in records:
+            if expected_sha1 != sha1s[version]:
                 raise errors.VersionedFileInvalidChecksum(version)
 
-    def get_sha1s(self, version_ids):
-        """Get the stored sha1 sums for the given revisions.
-
-        :param version_ids: The names of the versions to lookup
-        :return: a list of sha1s in order according to the version_ids
-        """
-        raise NotImplementedError(self.get_sha1s)
-
     def get_text(self, version_id):
         """Return version contents as a text string.
 
@@ -534,12 +527,36 @@
         self._backing_vf = backing_vf
         self.calls = []
 
+    def add_lines(self, key, parents, lines, parent_texts=None,
+        left_matching_blocks=None, nostore_sha=None, random_id=False,
+        check_content=True):
+        self.calls.append(("add_lines", key, parents, lines, parent_texts,
+            left_matching_blocks, nostore_sha, random_id, check_content))
+        return self._backing_vf.add_lines(key, parents, lines, parent_texts,
+            left_matching_blocks, nostore_sha, random_id, check_content)
+
+    def get_parent_map(self, keys):
+        self.calls.append(("get_parent_map", copy(keys)))
+        return self._backing_vf.get_parent_map(keys)
+
     def get_record_stream(self, keys, sort_order, include_delta_closure):
-        self.calls.append(("get_record_stream", keys, sort_order,
+        self.calls.append(("get_record_stream", list(keys), sort_order,
             include_delta_closure))
         return self._backing_vf.get_record_stream(keys, sort_order,
             include_delta_closure)
 
+    def get_sha1s(self, keys):
+        self.calls.append(("get_sha1s", copy(keys)))
+        return self._backing_vf.get_sha1s(keys)
+
+    def iter_lines_added_or_present_in_keys(self, keys, pb=None):
+        self.calls.append(("iter_lines_added_or_present_in_keys", copy(keys)))
+        return self._backing_vf.iter_lines_added_or_present_in_keys(keys, pb=pb)
+
+    def keys(self):
+        self.calls.append(("keys",))
+        return self._backing_vf.keys()
+
 
 class KeyMapper(object):
     """KeyMappers map between keys and underlying partitioned storage."""
@@ -745,10 +762,11 @@
                                   if not mpvf.has_version(p))
         # It seems likely that adding all the present parents as fulltexts can
         # easily exhaust memory.
-        present_parents = set(self.get_parent_map(needed_parents).keys())
         split_lines = osutils.split_lines
-        for record in self.get_record_stream(present_parents, 'unordered',
+        for record in self.get_record_stream(needed_parents, 'unordered',
             True):
+            if record.storage_kind == 'absent':
+                continue
             mpvf.add_version(split_lines(record.get_bytes_as('fulltext')),
                 record.key, [])
         for (key, parent_keys, expected_sha1, mpdiff), lines in\
@@ -819,7 +837,9 @@
         """Get the sha1's of the texts for the given keys.
 
         :param keys: The names of the keys to lookup
-        :return: a list of sha1s matching keys.
+        :return: a dict from key to sha1 digest. Keys of texts which are not
+            present in the store are not present in the returned
+            dictionary.
         """
         raise NotImplementedError(self.get_sha1s)
 
@@ -1051,9 +1071,9 @@
         sha1s = {}
         for prefix,suffixes, vf in self._iter_keys_vf(keys):
             vf_sha1s = vf.get_sha1s(suffixes)
-            for suffix, sha1 in zip(suffixes, vf.get_sha1s(suffixes)):
+            for suffix, sha1 in vf_sha1s.iteritems():
                 sha1s[prefix + (suffix,)] = sha1
-        return [sha1s[key] for key in keys]
+        return sha1s
 
     def insert_record_stream(self, stream):
         """Insert a record stream into this container.

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2008-06-11 04:20:16 +0000
+++ bzrlib/weave.py	2008-06-19 13:14:24 +0000
@@ -110,7 +110,7 @@
     def __init__(self, version, weave):
         """Create a WeaveContentFactory for version from weave."""
         ContentFactory.__init__(self)
-        self.sha1 = weave.get_sha1s([version])[0]
+        self.sha1 = weave.get_sha1s([version])[version]
         self.key = (version,)
         parents = weave.get_parent_map([version])[version]
         self.parents = tuple((parent,) for parent in parents)
@@ -765,7 +765,10 @@
 
     def get_sha1s(self, version_ids):
         """See VersionedFile.get_sha1s()."""
-        return [self._sha1s[self._lookup(v)] for v in version_ids]
+        result = {}
+        for v in version_ids:
+            result[v] = self._sha1s[self._lookup(v)]
+        return result
 
     def num_versions(self):
         """How many versions are in this weave?"""

